#include "poisson-sampling.hh"

#include <polymesh/std/hash.hh>
#include <unordered_map>

#include "libmorton/morton.h"

#include <iostream>

#include <bitset>

#include <queue>

std::pair<std::vector<std::vector<polymesh::vertex_handle>>, std::vector<tg::aabb3>> polymesh::create_hierarchical_poisson_sampling(
    polymesh::vertex_attribute<tg::pos3> const& position, float finest_sampling_radius, int num_layers)
{
    // DEBUGGING
    std::vector<tg::aabb3> debug_boxes;

    // 64-bit location codes, root node = 0b0..01
    // we can left-shift (ascend a layer) at most 21 times (21 * 3 = 63)
    // meaning we can shift 0 to 21 times -> 22 layers
    if (num_layers > 22)
    {
        // TODO proper warning?
        std::cout << "WARNING: sampling supports at most 22 layers, returning empty list!" << std::endl;
        return std::pair<std::vector<std::vector<polymesh::vertex_handle>>, std::vector<tg::aabb3>>();
    }

    // LATTICE TEMPLATES
    // to inspect neighborhood of a cell quickly here are precomputed offsets (sorted by decreasing query-ball-cell-coverage)
    // depending on cellsize-radius relation we either have to check 5x5x5 neighborhood or only 3x3x3
    // 117 cells:
    std::vector<tg::ivec3> const lattice_template_5
        = {{0, 0, 0},    {0, 0, 1},    {0, -1, 0},   {1, 0, 0},    {-1, 0, 0},  {0, 0, -1},   {0, 1, 0},   {1, 0, 1},   {1, 0, -1},   {-1, 0, 1},
           {0, -1, -1},  {-1, 1, 0},   {1, 1, 0},    {0, 1, -1},   {0, 1, 1},   {1, -1, 0},   {-1, -1, 0}, {0, -1, 1},  {-1, 0, -1},  {-1, -1, -1},
           {1, -1, 1},   {-1, 1, -1},  {1, -1, -1},  {-1, 1, 1},   {-1, -1, 1}, {1, 1, -1},   {1, 1, 1},   {0, -2, 0},  {0, 0, -2},   {-2, 0, 0},
           {0, 2, 0},    {0, 0, 2},    {2, 0, 0},    {-2, 0, 1},   {0, -2, 1},  {-1, 0, 2},   {0, -1, 2},  {-2, 1, 0},  {1, -2, 0},   {0, 2, 1},
           {0, 2, -1},   {0, -1, -2},  {-1, -2, 0},  {-1, 2, 0},   {1, 0, -2},  {0, 1, -2},   {0, -2, -1}, {1, 0, 2},   {-2, -1, 0},  {0, 1, 2},
           {-1, 0, -2},  {1, 2, 0},    {2, -1, 0},   {2, 0, -1},   {-2, 0, -1}, {2, 0, 1},    {2, 1, 0},   {-1, 2, -1}, {-1, 2, 1},   {1, -2, -1},
           {1, -2, 1},   {1, -1, -2},  {1, -1, 2},   {-2, -1, -1}, {1, 1, -2},  {-2, -1, 1},  {-2, 1, -1}, {-2, 1, 1},  {1, 1, 2},    {1, 2, -1},
           {-1, -2, -1}, {1, 2, 1},    {2, -1, -1},  {-1, -2, 1},  {2, -1, 1},  {-1, -1, -2}, {-1, -1, 2}, {-1, 1, -2}, {2, 1, -1},   {-1, 1, 2},
           {2, 1, 1},    {-2, 2, 0},   {0, 2, 2},    {0, 2, -2},   {-2, -2, 0}, {2, -2, 0},   {0, -2, 2},  {0, -2, -2}, {2, 0, -2},   {2, 0, 2},
           {-2, 0, 2},   {-2, 0, -2},  {2, 2, 0},    {1, 2, -2},   {-2, 2, -1}, {1, -2, -2},  {1, 2, 2},   {2, -2, -1}, {2, -2, 1},   {2, -1, -2},
           {-2, -1, 2},  {-2, 1, 2},   {-2, -1, -2}, {2, -1, 2},   {-2, -2, 1}, {1, -2, 2},   {-1, 2, 2},  {-2, 1, -2}, {-2, -2, -1}, {2, 1, -2},
           {-1, -2, 2},  {-1, -2, -2}, {-1, 2, -2},  {2, 1, 2},    {2, 2, -1},  {-2, 2, 1},   {2, 2, 1}};

    // when radius <= cellsize we only need to check the 3x3 neighborhood
    // 27 cells:
    std::vector<tg::ivec3> const lattice_template_3
        = {{0, 0, 0},  {-1, 0, 0},   {0, -1, 0},  {0, 0, -1},  {0, 0, 1},  {0, 1, 0},   {1, 0, 0},  {-1, -1, 0}, {-1, 0, -1},
           {-1, 0, 1}, {-1, 1, 0},   {0, -1, -1}, {0, -1, 1},  {0, 1, -1}, {0, 1, 1},   {1, -1, 0}, {1, 0, -1},  {1, 0, 1},
           {1, 1, 0},  {-1, -1, -1}, {-1, -1, 1}, {-1, 1, -1}, {-1, 1, 1}, {1, -1, -1}, {1, -1, 1}, {1, 1, -1},  {1, 1, 1}};

    auto const& m = position.mesh();
    auto const num_verts = m.vertices().size();

    typedef std::bitset<64> location_code;

    struct cubic_sparse_grid
    {
        // morton index into cell -> gives vertex handle if non-empty
        std::unordered_map<location_code, polymesh::vertex_handle> handle_in_cell;

        // cubic bounding box information & cell information
        tg::pos3 origin;

        // contrary to the paper we compute the following values *from* the sampling radius..
        // cell_extent = extent / num_cells_per_axis
        float cell_extent = 1.0f;
        int num_cells_per_axis = 1;
    };

    auto compute_cell_index = [](tg::pos3 const& pos, cubic_sparse_grid const& grid) {
        // compute the cell indices of 3d position w.r.t. some grid
        // ivec3 corresponds to a flooring
        // note: we allow this cell index to be out of bounds (i.e. negative or greater numcells)
        // that way one can check its neighbors' indices, which in turn may be in bounds again
        auto cell_indices = tg::ivec3((pos - grid.origin) / grid.cell_extent);
        return cell_indices;
    };

    // compute the morton index of a 3D position (e.g. some point)
    auto compute_morton_index = [](tg::pos3 const& pos, cubic_sparse_grid const& grid) {
        // compute the cell indices of 3d position w.r.t. some grid
        // ivec3 corresponds to a flooring
        auto cell_indices = tg::ivec3((pos - grid.origin) / grid.cell_extent);

        // make sure we are in bounds - should not compute morton index for points outside the grid
        TG_ASSERT(tg::min_element(cell_indices) >= 0);
        TG_ASSERT(tg::max_element(cell_indices) < grid.num_cells_per_axis);

        // TODO not sure if conversion to bitsets really necessary..
        return location_code(libmorton::morton3D_64_encode(cell_indices.x, cell_indices.y, cell_indices.z));
    };


    struct octree_node
    {
        // no need to store a location code here, since it's the key to this node in 'sparse_octree'
        // location_code = ..

        // existence flags for 8 potential children (octree nodes)
        // (for checking existence of child nodes before creating them or during traversal of the tree)
        std::bitset<8> child_active = 0b00000000; // initially no active children

        // center and equal sidelengths make up an "octant"
        tg::pos3 center;
        float half_sidelength = 1.0f;

        float sample_radius = 1.0f;

        // each octree node contains a cubic regular sparse grid
        cubic_sparse_grid sparse_grid;
    };

    // every octree node has a unique location code
    auto sparse_octree = std::unordered_map<location_code, octree_node>();

    // a query ball, created from a point's position and the radius of the node it wants to join
    struct query_ball
    {
        query_ball(tg::pos3 const& pos, float const radius) : pos(pos), radius(radius){};
        tg::pos3 pos;
        float radius = 1.0f;
    };

    // overlap test for query ball and octant (given as its location code)
    auto is_overlapping = [&](query_ball const& qb, location_code const code) {
        TG_ASSERT(sparse_octree.find(code) != sparse_octree.end());
        auto const& node = sparse_octree.at(code);
        auto const q = tg::abs(qb.pos - node.center);
        return (tg::max_element(q) < node.half_sidelength + qb.radius
                && (tg::min_element(q) < node.half_sidelength || qb.radius > tg::length(q - node.half_sidelength)));
    };

    // checks if a query ball conflicts with any previously inserted points in a grid
    // special case: query ball is outside - it doesn't lie in a cell, instead just check neighboring cells and make sure they exist first
    auto is_query_ball_conflicting = [&](query_ball const& qb, location_code code, bool const outside = false) {
        auto const& node = sparse_octree.at(code);

        // compute morton index of this position in grid
        auto const cell_index = compute_cell_index(qb.pos, node.sparse_grid);

        // note: the cell itself may be occupied, but it's a coarser cell, from a previous layer
        // the query ball stems from a *finer* level and does not care about cell occupancy in this node
        //
        // *no* need for:
        // auto const morton_index = libmorton::morton3D_64_encode(cell_index.x, cell_index.y, cell_index.z);
        // if (!outside && node.sparse_grid.handle_in_cell.find(morton_index) != node.sparse_grid.handle_in_cell.end())
        // reject.. <- would be wrong! don't reject on cell occupancy here!

        // check neighborhood
        {
            // here a 3x3 lattice may suffice, when outer cells of lattice not reachable by ball
            // (small qb.radius vs. large cellsizes)
            auto const& index_offsets = qb.radius < node.sparse_grid.cell_extent ? lattice_template_3 : lattice_template_5;

            for (auto const& offset : index_offsets)
            {
                auto const n_idx = cell_index + offset;

                // since query ball is outside neighbors may also be outside
                // only compute morton code and search for key if inside grid, else skip
                if (outside)
                    if (tg::min_element(n_idx) < 0 || tg::max_element(n_idx) >= node.sparse_grid.num_cells_per_axis)
                        continue;

                auto const m = location_code(libmorton::morton3D_64_encode(n_idx.x, n_idx.y, n_idx.z));

                // check if cell belonging to morton code contains a vertex
                auto const key_idx = node.sparse_grid.handle_in_cell.find(m);
                if (key_idx != node.sparse_grid.handle_in_cell.end())
                {
                    auto const nvh = (*key_idx).second;

                    // if too close, we have to reject this point
                    // note: using query ball's radius which is smaller or equal to this node's radius!
                    TG_ASSERT(qb.radius <= node.sample_radius);

                    if (tg::distance(position[nvh], qb.pos) < qb.radius)
                        return true;
                }
            }
        }

        // no conflicts with points in this grid found
        return false;
    };

    auto depth_from_location_code = [num_layers](location_code code) {
        if (0)
        {
            // idea: shifting >> 3 means going one layer up, hence depth is equal to number of possible shifts
            // once our code is equal to 1 we have reached the root and it's time to stop
            auto right_shift = 0;
            for (auto depth = 0u; depth < num_layers; ++depth)
            {
                if (code >> right_shift == 1)
                    return depth;
                right_shift += 3;
            }
        }
        else
        {
            auto i = 1UL;
            // TODO is this faster?
            return tg::u32(tg::log2(1.0f * code.to_ullong()) / 3);
        }
    };

    // creation of the octree's root node and its grid
    // contains the whole scene (and more, to be cubic and a multiple of our cellsize..)
    {
        octree_node root_node;

        // using aabb of mesh we create a cubic bounding box with cellsize according to our sampling radius
        auto const aabb = position.aabb();
        auto const extents = tg::abs(aabb.max - aabb.min);

        // given the leaf node sampling radius and number of levels - compute sampling radius of coarsest level (root node)
        // TODO for now assuming sampling radius is scaled by 2 each coarser layer
        root_node.sample_radius = tg::pow(2.0f, num_layers - 1) * finest_sampling_radius;

        std::cout << "root sampling radius: " << root_node.sample_radius << std::endl;

        // from the aabb and leaf radius we can determine all quantities of our grid:
        // TODO paper uses different (possibly wrong?) formula to compute the "poisson radius" from cell extent
        // we want the poisson radius to be sqrt(3) * sidelength (i.e. length of diagonal of cube) because then we can confidently reject points from occupied cells!
        root_node.sparse_grid.cell_extent = root_node.sample_radius / sqrt(3.0f); // TODO paper says * 2 ???
        root_node.sparse_grid.num_cells_per_axis = tg::ceil(tg::max_element(extents) / root_node.sparse_grid.cell_extent);
        root_node.sparse_grid.origin = aabb.min;

        // octant information for fast conflict checking of query balls later..
        root_node.half_sidelength = (root_node.sparse_grid.num_cells_per_axis * root_node.sparse_grid.cell_extent) / 2;
        root_node.center = aabb.min + root_node.half_sidelength;

        // root node has code 1
        sparse_octree[location_code(1)] = root_node;

        for (auto const& vh : m.vertices())
            TG_ASSERT(tg::max_element(tg::abs(position[vh] - root_node.center)) <= root_node.half_sidelength);
    }

    // for each point: its handle and a pointer to the octree node it would like to populate
    struct active_point
    {
        active_point(polymesh::vertex_handle vh, location_code tn) : vh(vh), target_node(tn){};
        polymesh::vertex_handle vh;
        location_code target_node;
    };

    auto sampled = 0u;

    // main section of the algorithm:
    // inserting points & creating octants and grids on-the-fly
    {
        // initializing queue with all points:
        auto active_queue = std::queue<active_point>(); // std::queue<active_point>();
        for (auto const vh : m.vertices())
            active_queue.emplace(vh, 1);

        // TODO paper 4.3 talks about accelerated search by "processing last layer of nodes that changed first"? what does that mean?
        // process all points until queue empty (all points processed or rejected due to depth limitations)
        while (!active_queue.empty())
        {
            auto cur_point = active_queue.front();
            active_queue.pop();

            // desired octree node
            TG_ASSERT(sparse_octree.find(cur_point.target_node) != sparse_octree.end());
            auto& node = sparse_octree[cur_point.target_node];

            // compute morton index of this position
            auto const cell_index = compute_cell_index(position[cur_point.vh], node.sparse_grid);
            auto const morton_index = location_code(libmorton::morton3D_64_encode(
                cell_index.x, cell_index.y, cell_index.z)); // compute_morton_index(position[cur_point.vh], node.sparse_grid);

            // the cell itself is not occupied..
            if (node.sparse_grid.handle_in_cell.find(morton_index) == node.sparse_grid.handle_in_cell.end())
            {
                // becomes true, when we find any point, that would be closer than node.sample_radius
                auto reject = false;

                // check neighbors in this grid first:
                // iterate over all offsets into all neighboring cells we may possibly touch
                auto const& index_offsets = lattice_template_5;
                for (auto const& offset : index_offsets)
                {
                    // TODO may land outside of grid, but should be fine and just return an unknown morton code..
                    auto const n_idx = cell_index + offset;
                    auto const m = location_code(libmorton::morton3D_64_encode(n_idx.x, n_idx.y, n_idx.z));

                    // neighbor cell contains a vertex
                    auto const key_idx = node.sparse_grid.handle_in_cell.find(m);
                    if (key_idx != node.sparse_grid.handle_in_cell.end())
                    {
                        auto const nvh = (*key_idx).second;

                        // if too close, we have to reject this point
                        if (tg::distance(position[nvh], position[cur_point.vh]) < node.sample_radius)
                        {
                            reject = true;
                            break;
                        }
                    }
                }

                // will descend from parent (if it exists)
                auto parent_code = cur_point.target_node >> 3;

                // only necessary if node has a parent (meaning it is not the root node, where no further grids would have to be checked)
                if (parent_code != 0)
                {
                    // point was not rejected in its grid - but we still have to check all other (possibly conflicting) grids..
                    //
                    // a query ball that stays fixed, no matter which octree layer we compare it to
                    auto const qb = query_ball(position[cur_point.vh], node.sample_radius);
                    // check if query ball fully contained in this octant:
                    // transform point position to local coordinate system
                    auto relative_point_pos = tg::abs(qb.pos - node.center);

                    // INSIDE TEST
                    if (tg::max_element(relative_point_pos + qb.radius) < node.half_sidelength)
                    {
                        // the query ball is fully inside this octant
                        // we only check its ancestors!
                        if (is_query_ball_conflicting(qb, parent_code, false))
                        {
                            reject = true;
                        }
                        else if (parent_code != 1) // if parent was root we can stop now
                        {
                            // TODO because paper says so:
                            // compute all ancestor codes (before parent) up to root first and traverse them *backwards* (root first)
                            auto ancestor_codes = std::vector<location_code>();

                            auto code = parent_code;
                            while (code != location_code(1))
                            {
                                code = code >> 3;
                                ancestor_codes.push_back(code);
                            }

                            // sanity check
                            TG_ASSERT(ancestor_codes.back() == 1);

                            // ..then descend from root down and stop before parent (because we've checked it already)
                            // iterate backwards through previously created list of codes and check their grids for conflicts
                            for (auto iter = ancestor_codes.rbegin(); iter != ancestor_codes.rend(); ++iter)
                                if (is_query_ball_conflicting(qb, *iter, false))
                                {
                                    reject = true;
                                    break;
                                }
                        }
                    }
                    // OVERLAP TESTS
                    else
                    {
                        auto const& parent_node = sparse_octree.at(parent_code);

                        for (auto i = 0u; i < 8; ++i)
                        {
                            // only if sibling exists
                            if (!parent_node.child_active[i])
                                continue;

                            auto const sibling_code = (parent_code << 3) | location_code(i);

                            // that's not a sibling but ourself, skip
                            if (sibling_code == cur_point.target_node)
                                continue;

                            {
                                // overlap test with sibling
                                if (is_overlapping(qb, sibling_code))
                                {
                                    if (is_query_ball_conflicting(qb, sibling_code, true))
                                    {
                                        reject = true;
                                        break;
                                    }
                                }
                            }
                        }

                        // TODO paper suggest to start from the root now?
                        if (!reject)
                        {
                            auto active_nodes = std::queue<location_code>({location_code(1)});

                            auto const parent_depth = depth_from_location_code(parent_code);

                            while (!active_nodes.empty())
                            {
                                auto node_code = active_nodes.front();
                                active_nodes.pop();

                                if (is_overlapping(qb, node_code))
                                {
                                    if (is_query_ball_conflicting(qb, node_code, true))
                                    {
                                        reject = true;
                                        break;
                                    }

                                    // add all existing children to queue if not too deep already
                                    // (ensures we aren't checking previously processed nodes again)
                                    // the equal sign is necessary to allow for our parent's siblings' children to be checked aswell
                                    // TODO checking these earlier (e.g. with siblings) may be more efficient than descending from root first..
                                    if (depth_from_location_code(node_code) <= parent_depth)
                                    {
                                        for (auto i = 0u; i < 8; ++i)
                                        {
                                            // only if child exists
                                            if (!sparse_octree.at(node_code).child_active[i])
                                                continue;

                                            auto const child_code = node_code << 3 | location_code(i);

                                            // making sure siblings and we are not again checked
                                            if (child_code == cur_point.target_node || child_code >> 3 == parent_code)
                                                continue;

                                            // push into queue
                                            active_nodes.push(child_code);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }


                // this point passed all tests, no conflicts, add it to it's desired grid and check next in queue
                if (!reject)
                {
                    node.sparse_grid.handle_in_cell[morton_index] = cur_point.vh;

                    // TODO just to show progress/that we are still running for larger point set..
                    if (0)
                    {
                        sampled++;
                        if (sampled % (1000 * 5) == 0)
                            std::cout << "sampled 5k more points -> " << sampled << std::endl;
                    }

                    continue;
                }
            }

            // if we made it until here that means the point was rejected and has to try again in a higher octree level
            {
                // first we check whether there even should be a finer level
                // if current depth is already max we must drop this point
                if (depth_from_location_code(cur_point.target_node) >= num_layers - 1)
                    continue;

                auto const node_corner = node.center - node.half_sidelength * tg::vec3::one;
                auto octant_index = tg::ivec3((position[cur_point.vh] - node_corner) / node.half_sidelength);

                // confirm that octant index is indeed valid
                // (could be that e.g. point lies outside octant)
                //
                // TODO numerical issues? instead clamp?
                if (0)
                {
                    TG_ASSERT(octant_index.x == 0 || octant_index.x == 1);
                    TG_ASSERT(octant_index.y == 0 || octant_index.y == 1);
                    TG_ASSERT(octant_index.z == 0 || octant_index.z == 1);
                }
                else if (tg::min_element(octant_index) < 0 || tg::max_element(octant_index) > 1)
                {
                    std::cout << "WARNING: octant index was outside bounds of octant - clamped it" << std::endl;
                    // TODO in VERY rare cases octant index slightly (how much exactly?) outside octant
                    // TODO clamping should fix that..
                    octant_index = tg::clamp(octant_index, tg::ivec3(0), tg::ivec3(1));
                }

                auto const morton_index = location_code(libmorton::morton3D_64_encode(octant_index.x, octant_index.y, octant_index.z));

                // does this child already exist?
                auto const child_index = morton_index.to_ullong();
                TG_ASSERT(child_index < 8);

                // a child's location code is its parent's location code shifted to the left..
                // ..OR-ed with its morton code in relation to parent (as computed above)
                auto const child_location_code = cur_point.target_node << 3 | morton_index;

                if (!node.child_active[child_index])
                {
                    TG_ASSERT(sparse_octree.find(child_location_code) == sparse_octree.end());

                    // create child
                    auto& newborn = sparse_octree[child_location_code];

                    TG_ASSERT(sparse_octree.find(child_location_code) != sparse_octree.end());

                    // tell octant child now exists
                    // TODO however this child is still empty - better inform parent later? -> won't traverse this empty child
                    // TODO would have to check whether a point insert is the first <-> "activating" that child
                    node.child_active[child_index] = 1;

                    // everything is halfed..
                    newborn.half_sidelength = node.half_sidelength / 2;
                    newborn.sample_radius = node.sample_radius / 2;

                    // TODO could simplify this computation but this is most readable..
                    // center is computed starting in corner of parent..
                    newborn.center = node_corner;
                    // ..move to corner of child
                    // (note how half_sidelength of parent node is full sidelength of a child)

                    newborn.center += node.half_sidelength * tg::vec3(octant_index);
                    // ..move to center of child
                    newborn.center += newborn.half_sidelength;

                    // uniform sparse grid origin is then:
                    newborn.sparse_grid.origin = newborn.center - newborn.half_sidelength * tg::vec3::one;
                    // extent of octant and its cells is just half of what the parent's
                    newborn.sparse_grid.cell_extent = node.sparse_grid.cell_extent / 2;
                    // ..meaning the number of cells stays constant!
                    newborn.sparse_grid.num_cells_per_axis = node.sparse_grid.num_cells_per_axis;


                    // DEBUGGING
                    debug_boxes.push_back(tg::aabb3(newborn.sparse_grid.origin, newborn.sparse_grid.origin + newborn.sparse_grid.cell_extent * newborn.sparse_grid.num_cells_per_axis));
                }

                // set this new location code and add to queue again
                cur_point.target_node = child_location_code;
                active_queue.push(cur_point);
            }
        }
    }

    // our sparse_octree is built and we can harvest the distributed points from each grid
    {
        // one list per layer:
        auto octree_layers = std::vector<std::vector<polymesh::vertex_handle>>();
        octree_layers.resize(num_layers);

        // starting from the root node, we traverse the tree breadth-first (queue)
        auto active_nodes = std::queue<location_code>({location_code(1)});

        while (!active_nodes.empty())
        {
            auto node_code = active_nodes.front();
            active_nodes.pop();

            // get node
            auto const& node = sparse_octree.at(node_code);

            // this saves memory - we don't have to queue the depth of a node but only its location code
            // (is also used before creating a child and should be cheap enough)
            auto depth = depth_from_location_code(node_code);

            // std::cout << node_code << " has depth " << depth << std::endl;

            // collect all points and add to corresponding layer in our returned vector
            for (auto const& kv : node.sparse_grid.handle_in_cell)
                octree_layers[depth].push_back(kv.second);

            // add all existing children to queue
            for (auto i = 0u; i < 8; ++i)
                if (node.child_active[i])
                    active_nodes.push(node_code << 3 | location_code(i));
        }

        return {octree_layers, debug_boxes};
    }
}

std::pair<std::vector<std::vector<polymesh::vertex_handle>>, std::vector<tg::aabb3>> polymesh::create_hierarchical_poisson_sampling(
    polymesh::vertex_attribute<tg::pos3> const& position, float leaf_splat_radius)
{
    // TODO ..
    auto const alpha = 2.0f;

    auto const scene_bb_diagonal = position.aabb().max - position.aabb().min;
    auto const scene_bb_diagonal_length = tg::length(scene_bb_diagonal);
    std::cout << "scene diagonal has length " << scene_bb_diagonal_length << ", leaf render radius: " << leaf_splat_radius << std::endl;

    // should lead to sufficiently populated levels
    auto num_lods = tg::floor(1 + tg::log2(scene_bb_diagonal_length * alpha / (2 * sqrt(3.0f) * leaf_splat_radius)));

    // note: sampling radius is render radius * sqrt(3) / alpha (!)
    auto leaf_sampling_radius = leaf_splat_radius * sqrt(3.0f) / alpha;     

    std::cout << "recommending " << num_lods << " lods with radii:" << std::endl;
    for (auto i = 0u; i < num_lods; ++i)
        std::cout << leaf_sampling_radius* tg::pow(2.0f, num_lods - i - 1) << std::endl;

    return create_hierarchical_poisson_sampling(position, leaf_sampling_radius, num_lods);
}
